/*
 *
 *   qrash: the second portable demo in the world
 *
 *   Copyright (C) 1997  Queue Members Group Art Division
 *   Coded by Mad Max / Queue Members Group (Mike Shirobokov)
 *   <mad_max@qmg.rising.ru>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 * 
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 */
#include <time.h>
#if defined(_XBOX) || defined(W32)
#include <io.h>
#elif  !defined (_WIN32_WCE)
#include <unistd.h>
#endif
//#include <fcntl.h>
#include "misc.h"
#include "sys.h"
#include "3d.h"
#include "poly2d.h"
#include "resource.h"
#include "music.h"
#if defined(__WATCOMC__) || defined (__DJGPP__)
#include <conio.h>
#endif

int sysSoundCard=0, sysVideoMode=3;
bool sysDebug = false;

TCollection<KBHandler*> kb_handlers;

void
sysRegisterKeyboardHandler( KBHandler* handler )
{
  kb_handlers.Insert( new KBHandler*(handler) );
}

void
sysUpdateFrame()
{
  KB key = sysGetKey();
  if( key ) {
    if( key == KB_ESC ) {
      Shutdown();
    }
    for( int i=0; i<kb_handlers.Count; i++ ) {
      if( (*kb_handlers[i])(key) ) break;
    }
  }
}

void
Startup( char* cmdline )
{
  srand( sysCurrentTime() );

  musNoSound = cmdline && strstr( (char*)cmdline, "-nosound" );
  bool saveconfig = cmdline && strstr( (char*)cmdline, "-saveconfig" );
  sysDebug = cmdline && strstr( (char*)cmdline, "-debug" );

  vidInitVideo();
  debug( "initializing music" );
  musInitMusic();
#if defined(_XBOX)
  debug( "initializing resources" );
  resInit( "D:\\"PROJECT_NAME".dat" );
#else
  resInit( PROJECT_NAME".dat" );
#endif

  debug( "running sysStartup" );
  sysStartup();

  if( saveconfig || !sysLoadConfig( PROJECT_NAME ".cfg" ) ) {
    debug( "choosing a sound card" );
    musChooseCard();
    sysVideoMode = vidChooseVideoMode();
  }

  if( saveconfig ) {
    sysSaveConfig( PROJECT_NAME ".cfg" );
  }
  debug( "initializing sound card" );
  musInitCard();
  debug( "setting video mode" );
  vidSetVideoMode(sysVideoMode);

  debug( "initializing 3d" );
  Init3D();
  sysResetTimer();
}

void
Shutdown()
{
  musStopMusic();
  musCloseMusic();
  vidCloseVideo();
  sysCleanup();
}

void
sysSaveConfig( char* filename ) {
  FILE* h = fopen( filename, "wb" );
  if( !h ) error( "Cannot write configuration file" );
  fwrite( &sysSoundCard, 1, sizeof(int), h );
  fwrite( &sysVideoMode, 1, sizeof(int), h );
  fwrite( &vidHiResMode, 1, sizeof(int), h );
  fclose(h);
}

bool
sysLoadConfig( char* filename )
{
  FILE* h = fopen( filename, "rb" );
  if( !h ) return false;
  fread( &sysSoundCard, 1, sizeof(int), h );
  fread( &sysVideoMode, 1, sizeof(int), h );
  fread( &vidHiResMode, 1, sizeof(int), h );
  fclose(h);
  return true;
}

void
error( char* msg ) {
  char str[256];
  sprintf( str, "\nFatal error: %s\n", msg );
  vidMessage(str);
  Shutdown();
}

size_t allocated=0, max_alloc=0;

void*
operator new( size_t size )
{
  return cmalloc(size);
}

void
operator delete( void* ptr )
{
  cfree(ptr);
}


void*
cmalloc( size_t size )
{
  size = ( size + sizeof(uint) - 1 ) / sizeof(uint) * sizeof(uint);
  char* ptr = (char*)malloc(size+sizeof(uint)*3);
  if( !ptr ) {
    char str[256];
    sprintf( str, "Not enough memory (%d allocated)", allocated );
    error( str );
  }
  ptr += sizeof(uint)*2;
  *(((uint*)ptr)-2) = 0x12345678;
  *(((uint*)ptr)-1) = size;
  *(uint*)(ptr+size) = 0x87654321;

  allocated+=size+sizeof(uint)*3; if( allocated>max_alloc) max_alloc=allocated;
//  printf( "cmalloc %d\n", size );
  return ptr;
}

void
cfree( void* ptr )
{
    if(ptr) {
	if( *(((uint*)ptr)-2) != 0x12345678 ) {
	    error( "Heap corrupted (top of the block)" );
	}
	uint size = *(((uint*)ptr)-1);
	if( *(uint*)((char*)ptr+size) != 0x87654321 ) {
	    error( "Heap corrupted (bottom of the block)" );
        }
        *(((uint*)ptr)-2) = 0;
        *(uint*)((char*)ptr+size) = 0;
        if( !size ) error( "freeing object of zero bytes" );
        free( (char*)ptr-sizeof(uint)*2 );
        allocated -= size;
//    printf( "cfree %d\n", size );
    }
}

int
sysGetNumber( char* str, int min, int max )
{
  printf( str );
  fflush( stdout );
  int c;
  do {
#if defined(__WATCOMC__) || defined (__DJGPP__)
    c = getch() - '0';
#else
    c = getchar() - '0';
#endif
    if( c == 27-48 ) error( "User abort" );
  } while( c < min || c > max );
  return c;
}

uint timer_offset = 0;

void sysResetTimer()
{
  timer_offset = 0;
}

uint sysTimer()
{
  return sysCurrentTime() - timer_offset;
}
